﻿<#
    .SYNOPSIS
    Determines whether Lync Server 2013/Skype for Business Server 2015 database updates are required.

    .DESCRIPTION
    Determines whether database updates are required. It performs the following checks:
    Detects whether the pool version is Lync Server 2013 or later
    Detects whether database mirroring is enabled
    Detects whether the primary and mirror SQL servers are online
    Detects whether the mirror server is principal for any databases
    Detects whether the local machine is a FE of the given pool
    Detects whether the CMS is on Lync Server 2013 or later
    And returns the following info:
    Returns list of local databases, their versions, and whether a database update is required (if the local server is a member of the pool)
    Returns list of backend databases, their versions, and whether a database update is required
    Returns list of CMS databases, their versions, and whether a database update is required

    .NOTES
    Version      	   		: 1.2 - See changelog at http://www.ehloworld.com/2609
    Wish list						: Better error trapping
    Rights Required			: TBD
    Sched Task Required	: No
    Lync Server Version	: 2013 through January 2014 Cumulative Update
    Author(s)    				: Dave Howe
                      : Pat Richard
    Email/Blog/Twitter	: daveh@microsoft.com @davehmcse
                        pat@innervation.com 	http://www.ehloworld.com @patrichard
    Dedicated Post			: http://www.ehloworld.com/2610
    Disclaimer   				: You running this script means you won't blame me if this breaks your stuff. This script is provided AS IS
                        without warranty of any kind. I disclaim all implied warranties including, without limitation, any implied
                        warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use
                        or performance of the sample scripts and documentation remains with you. In no event shall I be liable for
                        any damages whatsoever (including, without limitation, damages for loss of business profits, business
                        interruption, loss of business information, or other pecuniary loss) arising out of the use of or inability
                        to use the script or documentation.
    Acknowledgements 		:
    Assumptions					: ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
    Limitations					:
    Known issues				: script will abort if mirror sql server is principal for any lync database

    .EXAMPLE
    .\Get-CsDatabaseUpdateStatus.ps1

    Description
    -----------
    Runs the script. If there is only one pool in the environment, it will query that pool. If there are more pools, it will prompt for the pool FQDN to query.

    .EXAMPLE
    .\Get-CsDatabaseUpdateStatus.ps1 -PoolFqdn pool.contoso.com

    Description
    -----------
    Runs the script and queries the pool specified.
#>
#Requires -Version 2.0

[CmdletBinding(SupportsShouldProcess)]
param(
  # FQDN of the Lync pool to query
  [parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
  [string] $PoolFqdn,
	
  # Defines the location for any downloaded files. Defaults to "c:\_install". Additionally, log files generated by this script are located in a subfolder of TargetFolder called "logs". TargetFolder does not support paths with spaces, but does support non-hidden UNC paths.
  [parameter(ValueFromPipelineByPropertyName)]
  [string] $TargetFolder = "$env:SystemDrive\_Install"
)
[string] $ScriptVersion = '1.2'
[int] $Article = 2610
[string] $LogPath = "$TargetFolder\logs\$env:ComputerName" + ' {0:yyyy-MM-dd hh-mmtt}.log' -f (Get-Date)
[string] $LogDivider = '------------------------------'

#region functions
function Get-DBTestResults {
  [CmdletBinding(SupportsShouldProcess)]
  param(
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateSet('LocalDatabases', 'BackendDatabases', 'CmsDatabases')]
    [ValidateNotNullOrEmpty()]
    [string] $DBType,
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [string] $PoolFqdn,
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [string] $PrimarySqlFqdn
  )

  switch ($DBType) {
    LocalDatabases {
      if (((Get-CsPool -Identity $PoolFqdn).computers) -contains ((Get-CsComputer -Local).fqdn)){
        Write-Host "Local Databases: ($((Get-CsComputer -Local).fqdn))`n"
        $dbresults = @(Test-CsDatabase -LocalService | Where-Object {$_.SqlServerFqdn -match '(local)'})
      }
    }

    BackendDatabases {
      Write-Output -InputObject "Backend Databases: ($PrimarySqlFqdn)`n"
      $dbresults = @(Test-CsDatabase -ConfiguredDatabases -SqlServerFqdn $PrimarySqlFqdn)
    }

    CmsDatabases {
      if (-Not (Get-CsService -CentralManagement).version[0] -lt 6) {
        Write-Output -InputObject "Central Management Databases: ($((Get-CsManagementConnection).SqlServer))`n"
        $dbresults = @(Test-CsDatabase -CentralManagementDatabase -SqlServerFqdn ((Get-CsManagementConnection).SqlServer) -SqlInstanceName ((Get-CsManagementConnection).SqlInstance))
      }
    }
  }

  for ($int=0; $int -lt $dbresults.count; $int ++) {
    $databasename     += @([string]$dbresults[$int].databasename)
    $expectedversion  += @([string]$dbresults[$int].expectedversion)
    $installedversion += @([string]$dbresults[$int].installedversion)

    if (($databasename -match 'rtcdyn') -and ([string]$dbresults[$int].expectedversion -match '125.6.3') -and ([string]$dbresults[$int].installedversion -match '0.0.0')) {
      $updaterequired += @('no')
    } elseif (($databasename -match 'xds') -and ([string]$dbresults[$int].installedversion -match '10.13.2'))			{
      $updaterequired += @('no')
    } elseif ([string]$dbresults[$int].expectedversion -match [string]$dbresults[$int].installedversion)			{
      $updaterequired += @('no')
    } elseif ([string]$dbresults[$int].expectedversion -notmatch [string]$dbresults[$int].installedversion)			{
      $updaterequired += @('yes')
    }
  }

  $DatabasesData = @()

  for ($int=0; $int -lt $dbresults.count; $int ++) {
    $ThisObject = New-Object -TypeName PSObject
    Add-Member -InputObject $ThisObject -MemberType NoteProperty -Name DatabaseName -Value $DatabaseName[$int]
    Add-Member -InputObject $ThisObject -MemberType NoteProperty -Name ExpectedVersion -Value $ExpectedVersion[$int]
    Add-Member -InputObject $ThisObject -MemberType NoteProperty -Name InstalledVersion -Value $InstalledVersion[$int]
    Add-Member -InputObject $ThisObject -MemberType NoteProperty -Name UpdateRequired -Value $UpdateRequired[$int]
    $DatabasesData += $ThisObject
  }

  return $DatabasesData | Sort-Object -Property DatabaseName | Format-Table -Autosize
} # end function Get-DBTestResults

function Get-PoolData {
  [CmdletBinding(SupportsShouldProcess)]
  param(
    [parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [string] $PoolFqdn
  )
  if (!(Get-CsComputer -Pool $PoolFqdn))		{
    Write-Host "`nerror: pool not found in topology`n`n" -ForegroundColor red
    break
  }

  if ((Get-CsService -PoolFqdn $PoolFqdn).Version[0] -lt 6) {
    Write-Host "`nerror: incorrect server version (Lync Server 2013 required)`n`n" -ForegroundColor red
    break
  }

  $FrontEnds = @((Get-CsPool -Identity $PoolFqdn).Computers | Sort-Object)

  Write-Host "`n`n        pool fqdn:" $PoolFqdn

  if ($FrontEnds.Count -eq 1)		{
    Write-Host '   front end fqdn:' $FrontEnds[0]
  }	else		{
    Write-Host '  front end fqdns:' $FrontEnds[0]
    for ($int = 1; $int -lt $FrontEnds.Count; $int ++)			{
      Write-Host '                  ' $FrontEnds[$int]
    }
  }

  return $FrontEnds
} # end function Get-PoolData

function Get-UpdateInfo {
  <#
      .SYNOPSIS
      Queries an online XML source for version information to determine if a new version of the script is available.

      .DESCRIPTION
      Queries an online XML source for version information to determine if a new version of the script is available.

      .NOTES
      Version               : 1.5 - See changelog at https://ucunleashed.com/3168 for fixes & changes introduced with each version
      Wish list             : Better error trapping
      Rights Required       : N/A
      Sched Task Required   : No
      Lync/Skype4B Version  : N/A
      Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
      Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
      Donations             : https://www.paypal.me/PatRichard
      Dedicated Post        : https://ucunleashed.com/3168
      Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This script/function 
                              is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including, without limitation, 
                              any implied warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use 
                              or performance of the sample scripts and documentation remains with you. In no event shall author(s) be held liable for 
                              any damages whatsoever (including, without limitation, damages for loss of business profits, business interruption, loss 
                              of business information, or other pecuniary loss) arising out of the use of or inability to use the script or 
                              documentation. Neither this script/function, nor any part of it other than those parts that are explicitly copied from 
                              others, may be republished without author(s) express written permission. Author(s) retain the right to alter this 
                              disclaimer at any time. For the most up to date version of the disclaimer, see https://ucunleashed.com/code-disclaimer.
      Acknowledgements      : Reading XML files 
                            http://stackoverflow.com/questions/18509358/how-to-read-xml-in-powershell
                            http://stackoverflow.com/questions/20433932/determine-xml-node-exists
                          : TLS 1.2 fixup (via @GreigInSydney)
                            https://github.com/chocolatey/choco/wiki/Installation#installing-with-restricted-tls
      Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
      Limitations           : 
      Known issues          : 

      .EXAMPLE
      Get-UpdateInfo -Article 123

      Description
      -----------
      Runs function to check for updates to article/script 123.

      .INPUTS
      None. You cannot pipe objects to this script.
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpUri = 'https://ucunleashed.com/3168')]
  param (
    # Specifies the blog article number. This is used to check the XML file for version info.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [int] $Article,
    
    # Path to the XML file that contains the update information.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [string] $XmlPath = 'https://ucunleashed.com/downloads/version.xml'
  )
  [bool] $HasInternetAccess = ([Activator]::CreateInstance([Type]::GetTypeFromCLSID([Guid]'{DCB00C01-570F-4A9B-8D69-199FDBA5723B}')).IsConnectedToInternet)
  if ($HasInternetAccess){
    Write-Log -Level Info -Message 'Performing update check'
    # ------------------ TLS 1.2 fixup
    $SecurityProtocolSettingsOriginal = [Net.ServicePointManager]::SecurityProtocol
    Write-Log -Level Info -Message "Original security protocol settings: $SecurityProtocolSettingsOriginal" -NoConsole
    try {
      # Set TLS 1.2 (3072). Use integers because the enumeration values for TLS 1.2 won't exist in .NET 4.0, even though they are 
      # addressable if .NET 4.5+ is installed (.NET 4.5 is an in-place upgrade).
      Write-Log -Level Info -Message 'New security protocol settings: Tls12 (3072)' -NoConsole -Indent 1
      [Net.ServicePointManager]::SecurityProtocol = 3072
      # Need to error check here to ensure it was actually changed.
    } catch {
      Write-Log -Level Warn -Message 'Unable to set PowerShell to use TLS 1.2 due to old .NET Framework installed.' -Indent 1
    }
    # ------------------ end TLS 1.2 fixup
    Write-Log -Level Info -Message "Retrieving update info for article $Article from $XmlPath" -NoConsole -Indent 1
    [xml] $xml = (New-Object -TypeName System.Net.WebClient).DownloadString($XmlPath)
    # Reinstate original SecurityProtocol settings from before the TLS 1.2 fixup
    Write-Log -Level Info -Message "Restoring original security protocol settings to $SecurityProtocolSettingsOriginal" -NoConsole -Indent 1
    [Net.ServicePointManager]::SecurityProtocol = $SecurityProtocolSettingsOriginal    
    try{
      [object] $ArticleXml = Select-Xml -Xml $xml -XPath "//article[@id='$($article)']"
      [string] $Ga = $ArticleXml.Node.Version.Trim()
      Write-Log -Level Info -Message "GA version according to XML is $Ga" -NoConsole -Indent 1
    
      [string] $Title = $ArticleXml.Node.Title.Trim()
      if (-Not ([string] $DownloadUrl = $ArticleXml.Node.DownloadUrl.Trim())){
        [string] $DownloadUrl = "https://ucunleashed.com/$article#downloads"
      }
      [string] $changelog = 'This version includes: ' + $ArticleXml.Node.Changelog.Trim()

      if ($Ga -gt $ScriptVersion){
        Write-Log -Level Warn -Message "Outdated version. Version $Ga is latest version. Prompting user to upgrade." -NoConsole
        if ((New-Popup -Message "Version $Ga is now available. $changelog `n`nWould you like to download it?" -Title "New version of $Title available!" -Buttons YesNo -Icon Question) -eq 6){
          Write-Log -Level Info -Message "User elected to download version $Ga." -NoConsole -Indent 1
          Write-Log -Level Info -Message "Opening Internet Explorer and going to $DownloadUrl" -NoConsole -Indent 1
          Start-Process -FilePath $DownloadUrl
          Write-Log -Level Warn -Message "Script is exiting. Please start the new version of the script once you've downloaded it." -NoLog
          Stop-Script
        }else{
          Write-Log -Level Warn -Message "User elected NOT to download version $Ga." -Indent 1 -NoConsole
        }
      }elseif ($Ga -eq $ScriptVersion){
        Write-Log -Level Info -Message "Current version. Version $Ga is latest version." -NoConsole
          }elseif (-Not ($Ga)) {
        Write-Log -Level Warn -Message 'No version info available' -NoConsole
      }
    }
    catch{}
  }else{
    Write-Log -Level Warn -Message 'Unable to check online for update info.' -NoConsole
  }
  Write-Log -Level Info -Message 'Completed update check' -NoConsole -Indent 1
} # end function Get-UpdateInfo

function Get-SqlServerData {
  [CmdletBinding(SupportsShouldProcess)]
  param(
    [parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [string] $PoolFqdn
  )
  $Services = @(Get-CsService -PoolFqdn $PoolFqdn)

  foreach ($Service in $Services)		{
    if ($Service.role -contains 'UserServer')			{
      if ($Service.UserDatabase) {
        $PrimarySqlFqdn = $Service.UserDatabase.Split(':')[1]
      }
      if ($Service.MirrorUserDatabase) {
        $MirrorSqlFqdn = $Service.MirrorUserDatabase.Split(':')[1]
      }
      } elseif ($Service.role -contains 'PersistentChat') {
        if ($Service.PersistentChatDatabase) {
          $PrimarySqlFqdn = $Service.PersistentChatDatabase.Split(':')[1]
        }
      if ($Service.MirrorPersistentChatDatabase) {
        $MirrorSqlFqdn = $Service.MirrorPersistentChatDatabase.Split(':')[1]
      }
    } elseif ($Service.role -contains 'Registrar') {
      if ($Service.MonitoringDatabase) {
        $PrimarySqlFqdn = $Service.MonitoringDatabase.split(':')[1]
      }
      if ($Service.MirrorMonitoringDatabase) {
        $MirrorSqlFqdn = $Service.MirrorMonitoringDatabase.split(':')[1]
      }
    }
  }

  if ($PrimarySqlFqdn -ne $null)		{
    if(!(Test-Connection -ComputerName $PrimarySqlFqdn -BufferSize 16 -Count 1 -ErrorAction SilentlyContinue -Quiet))			{
      Write-Host "`n`nerror: Primary sql server $PrimarySqlFqdn is down`n`n" -ForegroundColor red
      break
    }
  }

  if ($MirrorSqlFqdn -ne $null)		{
    if(!(Test-Connection -ComputerName $MirrorSqlFqdn -BufferSize 16 -Count 1 -ErrorAction SilentlyContinue -Quiet))			{
      Write-Host "`n`nerror: Mirror sql server $MirrorSqlFqdn is down`n`n" -ForegroundColor red
      break
    }
  }

  if ($PrimarySqlFqdn -ne $null)		{
    Write-Host ' primary sql fqdn:' $PrimarySqlFqdn -NoNewLine
    $Principal = @(Get-CsDatabaseMirrorState -PoolFqdn $PoolFqdn | Where-Object {$_.StateOnPrimary -match 'Principal'} | Select-Object -Property DatabaseName -ExpandProperty DatabaseName)
    if ($Principal)			{
      Write-Host " ($Principal)"
    }		else			{
      Write-Host
    }
  }

  if ($MirrorSqlFqdn -ne $null)		{
    Write-Host '  mirror sql fqdn:' $MirrorSqlFqdn -NoNewLine
    $Principal = @(Get-CsDatabaseMirrorState -PoolFqdn $PoolFqdn | Where-Object {$_.StateOnMirror -match 'Principal'} | Select-Object -Property DatabaseName -ExpandProperty DatabaseName)
    if ($Principal)			{
      Write-Host " ($Principal)`n"
      Write-Host "error: Failover databases to primary sql server`n`n" -ForegroundColor red
      break
    }		else			{
      Write-Host
    }
  }
  return $PrimarySqlFqdn, $MirrorSqlFqdn
} # end function Get-SQLServerData

function New-Popup {
  <#
      .Synopsis
      Display a Popup Message

      .Description
      This command uses the Wscript.Shell PopUp method to display a graphical message
      box. You can customize its appearance of icons and buttons. By default the user
      must click a button to dismiss but you can set a timeout value in seconds to
      automatically dismiss the popup.

      The command will write the return value of the clicked button to the pipeline:
      OK     = 1
      Cancel = 2
      Abort  = 3
      Retry  = 4
      Ignore = 5
      Yes    = 6
      No     = 7

      If no button is clicked, the return value is -1.

      .Example
      PS C:\> new-popup -message "The update script has completed" -title "Finished" -time 5

      This will display a popup message using the default OK button and default
      Information icon. The popup will automatically dismiss after 5 seconds.

      .Notes
      Last Updated: April 8, 2013
      Version     : 1.0

      .Inputs
      None

      .Outputs
      integer

      Null   = -1
      OK     = 1
      Cancel = 2
      Abort  = 3
      Retry  = 4
      Ignore = 5
      Yes    = 6
      No     = 7
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging)]
  Param (
    # Message to be displayed in the popup message dialog box.
    [Parameter(Position = 0, Mandatory, HelpMessage = 'Enter a message for the popup')]
    [ValidateNotNullorEmpty()]
    [string] $Message,

    # Title of the popup dialog box.
    [Parameter(Position = 1, Mandatory, HelpMessage = 'Enter a title for the popup')]
    [ValidateNotNullorEmpty()]
    [string] $Title,

    # If defined, automatically dismisses the dialog box after the specified number of seconds. Defaults to 0, which requires a mouse click.
    [Parameter(Position = 2, Mandatory, HelpMessage = 'How many seconds to display? Use 0 require a button click.')]
    [ValidateScript({$_ -ge 0})]
    [int] $Time = 0,

    # Type of buttons to display.
    [Parameter(Position = 3, Mandatory, HelpMessage = 'Enter a button group')]
    [ValidateNotNullorEmpty()]
    [ValidateSet('OK','OKCancel','AbortRetryIgnore','YesNo','YesNoCancel','RetryCancel')]
    [string] $Buttons = 'OK',

    # Icon set to use in dialog box.
    [Parameter(Position = 4, Mandatory, HelpMessage = 'Enter an icon set')]
    [ValidateNotNullorEmpty()]
    [ValidateSet('Stop','Question','Exclamation','Information' )]
    [string] $Icon = 'Information'
  ) # end of param block

  #convert buttons to their integer equivalents
  Switch ($Buttons) {
    'OK'               {$ButtonValue = 0}
    'OKCancel'         {$ButtonValue = 1}
    'AbortRetryIgnore' {$ButtonValue = 2}
    'YesNo'            {$ButtonValue = 4}
    'YesNoCancel'      {$ButtonValue = 3}
    'RetryCancel'      {$ButtonValue = 5}
  }

  #set an integer value for Icon type
  Switch ($Icon) {
    'Stop'        {$iconValue = 16}
    'Question'    {$iconValue = 32}
    'Exclamation' {$iconValue = 48}
    'Information' {$iconValue = 64}
  }

  #create the COM Object
  Try {
    $wshell = New-Object -ComObject Wscript.Shell -ErrorAction Stop
    #Button and icon type values are added together to create an integer value
    $wshell.Popup($Message,$Time,$Title,$ButtonValue+$iconValue)
  }

  Catch {
    #You should never really run into an exception in normal usage
    Write-Log -Level Warn -Message 'Failed to create Wscript.Shell COM object'
    Write-Log -Level Warn -Message $_.exception.message
  }
} # end function New-Popup

function Set-ModuleStatus {
  <#
      .SYNOPSIS
      Imports a specified PowerShell module, with error checking.

      .DESCRIPTION
      Imports a specified PowerShell module, with error checking.

      .NOTES
      Version               : 1.6
      Wish list             : Better error trapping
      Rights Required       : None
      Sched Task Required   : No
      Lync/Skype4B Version  : N/A
      Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
      Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
      Donations             : https://www.paypal.me/PatRichard
      Dedicated Post        :
      Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This script/function 
                              is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including, without limitation, 
                              any implied warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use 
                              or performance of the sample scripts and documentation remains with you. In no event shall author(s) be held liable for 
                              any damages whatsoever (including, without limitation, damages for loss of business profits, business interruption, loss 
                              of business information, or other pecuniary loss) arising out of the use of or inability to use the script or 
                              documentation. Neither this script/function, nor any part of it other than those parts that are explicitly copied from 
                              others, may be republished without author(s) express written permission. Author(s) retain the right to alter this 
                              disclaimer at any time. For the most up to date version of the disclaimer, see https://ucunleashed.com/code-disclaimer.
      Acknowledgements      :
      Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
      Limitations           :
      Known issues          : None yet, but I'm sure you'll find some!

      .LINK

      .EXAMPLE
      .\

      Description
      -----------


      .INPUTS
      None. You cannot pipe objects to this script.
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpUri = 'https://ucunleashed.com/938')]
  param  (
    [parameter(ValueFromPipeline, ValueFromPipelineByPropertyName, Mandatory, HelpMessage = 'No module name specified!')]
    [ValidateNotNullOrEmpty()]
    [string] $Name
  ) # end of param block
  PROCESS{
    # Executes once for each pipeline object
    # the $_ variable represents the current input object
    if (-Not (Get-Module -name "$name")) {
      if (Get-Module -ListAvailable | Where-Object Name -eq "$name") {
        Import-Module -Name "$name"
        # module was imported
        # return $true
      } 
    }
  } # end PROCESS
} # end function Set-ModuleStatus

function Test-IsAdmin {
  <#
      .SYNOPSIS
      Determines if the user running the script is a local admin.

      .DESCRIPTION
      Determines if the user running the script is a local admin.

      .NOTES
      Version               : 1.0
      Wish list             : Better error trapping
      Rights Required       : Local administrator on server
      Sched Task Required   : No
      Lync/Skype4B Version  : N/A
      Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
      Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
      Donations             : https://www.paypal.me/PatRichard
      Dedicated Post        :
      Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This script/function 
                              is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including, without limitation, 
                              any implied warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use 
                              or performance of the sample scripts and documentation remains with you. In no event shall author(s) be held liable for 
                              any damages whatsoever (including, without limitation, damages for loss of business profits, business interruption, loss 
                              of business information, or other pecuniary loss) arising out of the use of or inability to use the script or 
                              documentation. Neither this script/function, nor any part of it other than those parts that are explicitly copied from 
                              others, may be republished without author(s) express written permission. Author(s) retain the right to alter this 
                              disclaimer at any time. For the most up to date version of the disclaimer, see https://ucunleashed.com/code-disclaimer.
      Acknowledgements      :
      Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
      Limitations           :
      Known issues          : None yet, but I'm sure you'll find some!

      .EXAMPLE
      .\

      Description
      -----------


      .INPUTS
      None. You cannot pipe objects to this script.
  #>
  [CmdletBinding(SupportsPaging)]
  param(
  ) # end of param block
  $principal = New-Object -TypeName System.Security.Principal.WindowsPrincipal -ArgumentList ([Security.Principal.WindowsIdentity]::GetCurrent())
  $principal.IsInRole([Security.Principal.WindowsBuiltInRole]::Administrator)
} # end function Test-IsAdmin

function Write-Log {
  <#
      .SYNOPSIS
      Extensive function to write data to either the console screen, a log file, and/or a Windows event log.

      .DESCRIPTION
      Extensive function to write data to either the console screen, a log file, and/or a Windows event log. Data can be written as info, warning, error, and includes indentation, time stamps, etc.

      .NOTES
      Version               : 3.2
      Wish list             : Better error trapping
      Rights Required       : Local administrator on server if writing to event log(s)
      Sched Task Required   : No
      Lync/Skype4B Version  : N/A
      Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
      Email/Blog/Twitter		: pat@innervation.com 	https://www.ucunleashed.com @patrichard
      Donations             : https://www.paypal.me/PatRichard
      Dedicated Post        : http://poshcode.org/6894
      Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This script/function 
                              is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including, without limitation, 
                              any implied warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use 
                              or performance of the sample scripts and documentation remains with you. In no event shall author(s) be held liable for 
                              any damages whatsoever (including, without limitation, damages for loss of business profits, business interruption, loss 
                              of business information, or other pecuniary loss) arising out of the use of or inability to use the script or 
                              documentation. Neither this script/function, nor any part of it other than those parts that are explicitly copied from 
                              others, may be republished without author(s) express written permission. Author(s) retain the right to alter this 
                              disclaimer at any time. For the most up to date version of the disclaimer, see https://ucunleashed.com/code-disclaimer.
      Acknowledgements      : Based on an original function by Any Arismendi, along with updates by others
                              http://poshcode.org/2566
                            
                              Test for log names and sources
                              http://powershell.com/cs/blogs/tips/archive/2013/06/10/testing-event-log-names-and-sources.aspx
                            
                              Writing to different event logs and sources registered to a single event log
                              http://social.technet.microsoft.com/Forums/en-US/winserverpowershell/thread/e172f039-ce88-4c9f-b19a-0dd6dc568fa0/
      Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
      Limitations           : Writing to event logs requires admin rights
      Known issues          :

      .EXAMPLE
      .\

      Description
      -----------


      .INPUTS
      System.String. You cannot pipe objects to this script.

      .OUTPUTS
      System.String
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpUri = 'http://poshcode.org/6894')]
  param(
    # The type of message to be logged. Alias is 'type'.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateSet('Error', 'Warn', 'Info', 'Verbose')]
    [ValidateNotNullOrEmpty()]
    [string] $Level = 'Info',

    # The message to be logged.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName, Mandatory, HelpMessage = 'No message specified.')]
    [ValidateNotNullOrEmpty()]
    [string] $Message,

    # Specifies that $message should not the sent to the log file.
    [Parameter(ValueFromPipelineByPropertyName)]
    [switch] $NoLog,

    # Specifies to not display the message to the console.
    [Parameter(ValueFromPipelineByPropertyName)]
    [switch] $NoConsole,

    # The number of spaces to indent the message in the log file.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateRange(1,30)]
    [ValidateNotNullOrEmpty()]
    [int] $Indent = 0,

    # Specifies what color the text should be be displayed on the console. Ignored when switch 'NoConsoleOut' is specified.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateSet('Black', 'DarkMagenta', 'DarkRed', 'DarkBlue', 'DarkGreen', 'DarkCyan', 'DarkYellow', 'Red', 'Blue', 'Green', 'Cyan', 'Magenta', 'Yellow', 'DarkGray', 'Gray', 'White')]
    [ValidateNotNullOrEmpty()]
    [String] $ConsoleForeground = 'White',

    # Existing log file is deleted when this is specified. Alias is 'Overwrite'.
    [Parameter(ValueFromPipelineByPropertyName)]
    [Switch] $Clobber,

    # The name of the system event log, e.g. 'Application'. The Skype for Business log is still called 'Lync Server'. Note that writing to the system event log requires elevated permissions.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateSet('Application', 'System', 'Security', 'Lync Server', 'Microsoft Office Web Apps')]
    [ValidateNotNullOrEmpty()]
    [String] $EventLogName,

    # The name to appear as the source attribute for the system event log entry. This is ignored unless 'EventLogName' is specified.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [String] $EventSource = $([IO.FileInfo] $MyInvocation.ScriptName).Name,

    # The ID to appear as the event ID attribute for the system event log entry. This is ignored unless 'EventLogName' is specified.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateRange(1,65535)]
    [ValidateNotNullOrEmpty()]
    [int] $EventID = 1,

    # The text encoding for the log file. Default is ASCII.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateSet('Unicode','Byte','BigEndianUnicode','UTF8','UTF7','UTF32','ASCII','Default','OEM')]
    [ValidateNotNullOrEmpty()]
    [String] $LogEncoding = 'ASCII'
        
    # Destination folder for log files
    #[Parameter(ValueFromPipelineByPropertyName)]
    #[ValidateNotNullOrEmpty()]
    #[string] $LogPath = "$TargetFolder\logs\$env:ComputerName" + " {0:yyyy-MM-dd hh-mmtt}.log" -f (Get-Date),

    # Divider line to be used to separate sections in the log file
    #[Parameter(ValueFromPipelineByPropertyName, ParameterSetName = 'Divider')]
    #[ValidateNotNullOrEmpty()]
    #[string] $LogDivider = '+------------------------------+'
  ) # end of param block
  BEGIN{
    [string]$LogFolder = Split-Path -Path $LogPath -Parent
    if (-not (Test-Path -Path $LogFolder)){
      $null = New-Item -Path $LogFolder -ItemType Directory
    }
  } # end BEGIN
  PROCESS{
    try {
      $Message = $($Message.trim())
      $msg = '{0} : {1} : {2}{3}' -f (Get-Date -Format 'yyyy-MM-dd HH:mm:ss'), $Level.ToUpper(), ('  ' * $Indent), $Message
      if (-not ($NoConsole)){
        switch ($Level) {
          'Error' {$Host.UI.WriteErrorLine("$Message")}
          'Warn' {Write-Warning -Message $Message}
          'Info' {Write-Host $Message -ForegroundColor $ConsoleForeground}
          'Verbose' {Write-Verbose -Message $Message}
        }
      }
      if (-not ($NoLog)){
        if ($Clobber) {
          $msg | Out-File -FilePath $LogPath -Encoding $LogEncoding -Force
        } else {
          $msg | Out-File -FilePath $LogPath -Encoding $LogEncoding -Append
        }
      }
      if ($EventLogName) {
        if (-not $EventSource) {
          [string] $EventSource = $([IO.FileInfo] $MyInvocation.ScriptName).Name
        }

        if(-not [Diagnostics.EventLog]::SourceExists($EventSource)) {
          [Diagnostics.EventLog]::CreateEventSource($EventSource, $EventLogName)
        }

        switch ($Level) {
          'Error' {$EntryType = 'Error'}
          'Warn'  {$EntryType = 'Warning'}
          'Info'  {$EntryType = 'Information'}
          'Verbose' {$EntryType = 'Information'}
          Default  {$EntryType = 'Information'}
        }
        Write-EventLog -LogName $EventLogName -Source $EventSource -EventId 1 -EntryType $EntryType -Message $Message
      }
      $msg = ''
    } # end try
    catch {
      Throw "Failed to create log entry in: '$LogPath'. The error was: '$_'."
    } # end catch
  } # end PROCESS
  END{} # end END
} # end function Write-Log
#endregion functions

#region LogHeader
Write-Log -Level Info -Message "Version           : $ScriptVersion" -NoConsole
Write-Log -Level Info -Message "User              : $env:UserDomain\$env:UserName" -NoConsole
Write-Log -Level Info -Message "Local Admin       : $(Test-IsAdmin)" -NoConsole
Write-Log -Level Info -Message "Command           : $($MyInvocation.Line)" -NoConsole
Write-Log -Level Info -Message "Server            : $env:ComputerName" -NoConsole
Write-Log -Level Info -Message "Culture           : $((Get-Culture).Name) `"$((Get-Culture).Displayname)`"" -NoConsole
Write-Log -Level Info -Message "Log path          : $LogPath" -NoConsole
Write-Log -Level Info -Message $LogDivider -NoConsole
#endregion LogHeader

Get-UpdateInfo -Article $Article

#region main
if ((! $PoolFqdn) -and ((Get-CsService -Registrar | Measure-Object).count -eq 1)){
  $PoolFqdn = (Get-CsService -Registrar).PoolFqdn
}
Write-Host

if (! $PoolFqdn){
  $PoolFqdn = Read-Host -Prompt 'server/pool FQDN'
}

Write-Host "`n`nChecking..." -NoNewLine

if ($PoolData = @(Get-PoolData -PoolFqdn $PoolFqdn)) {
  $FrontEnds = $PoolData
}

if ($SqlServerData = @(Get-SqlServerData -PoolFqdn $PoolFqdn)) {
  $PrimarySqlFqdn = $SqlServerData[0]
  $MirrorSqlFqdn = $SqlServerData[1]
}

Write-Host "`n`n"

Get-DBTestResults -DBType LocalDatabases -PoolFqdn $PoolFqdn
Get-DBTestResults -DBType BackendDatabases -PoolFqdn $PoolFqdn -PrimarySqlFqdn $PrimarySqlFqdn
Get-DBTestResults -DBType CmsDatabases

#end region main
# SIG # Begin signature block
# MIIcdgYJKoZIhvcNAQcCoIIcZzCCHGMCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUn/oktwIvhRIDYB2YmfW8UAcv
# noygghelMIIFLjCCBBagAwIBAgIQAWzpe+2ITyY5e1fzGr5MLDANBgkqhkiG9w0B
# AQsFADByMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYD
# VQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYDVQQDEyhEaWdpQ2VydCBTSEEyIEFz
# c3VyZWQgSUQgQ29kZSBTaWduaW5nIENBMB4XDTE3MTAyNTAwMDAwMFoXDTE4MTAz
# MDEyMDAwMFowazELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAk1JMRkwFwYDVQQHExBT
# dGVybGluZyBIZWlnaHRzMRkwFwYDVQQKExBJbm5lcnZhdGlvbiwgTExDMRkwFwYD
# VQQDExBJbm5lcnZhdGlvbiwgTExDMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIB
# CgKCAQEA24OX3nR2gF8WQ+Te3uJSrV3kOUoxUkczclPxyRXzsfxmkOu4sGU1EeCk
# 5R1fBk68dXuvFbdPUJcKTF2l8kFj6F5a6Ujr5OB9NCBurl96Y+F99RTmYMH0IClt
# YRuWfN1TqNgn9Cn0+uoVwQs8VLFV27y3nIu/ry+eHF5UPvyAnYxFnKkehV4xmibr
# gezzm23MVar0F4p95O8VoJypfvY+gWiPutkrtyRT/cIAcS+7F0F5ZGqEC6H1N8DY
# dXW0NiBF8F5/lWp9K2IGAzVv/vlEXBGTtM9LP5FrEGXr6CeJrmuPfom73ugqe851
# YFxeewyDfVs7fu5tRKsDz0N1EH2FdwIDAQABo4IBxTCCAcEwHwYDVR0jBBgwFoAU
# WsS5eyoKo6XqcQPAYPkt9mV1DlgwHQYDVR0OBBYEFFRsPiAppJf9S2b8T7Edzfbz
# r5EFMA4GA1UdDwEB/wQEAwIHgDATBgNVHSUEDDAKBggrBgEFBQcDAzB3BgNVHR8E
# cDBuMDWgM6Axhi9odHRwOi8vY3JsMy5kaWdpY2VydC5jb20vc2hhMi1hc3N1cmVk
# LWNzLWcxLmNybDA1oDOgMYYvaHR0cDovL2NybDQuZGlnaWNlcnQuY29tL3NoYTIt
# YXNzdXJlZC1jcy1nMS5jcmwwTAYDVR0gBEUwQzA3BglghkgBhv1sAwEwKjAoBggr
# BgEFBQcCARYcaHR0cHM6Ly93d3cuZGlnaWNlcnQuY29tL0NQUzAIBgZngQwBBAEw
# gYQGCCsGAQUFBwEBBHgwdjAkBggrBgEFBQcwAYYYaHR0cDovL29jc3AuZGlnaWNl
# cnQuY29tME4GCCsGAQUFBzAChkJodHRwOi8vY2FjZXJ0cy5kaWdpY2VydC5jb20v
# RGlnaUNlcnRTSEEyQXNzdXJlZElEQ29kZVNpZ25pbmdDQS5jcnQwDAYDVR0TAQH/
# BAIwADANBgkqhkiG9w0BAQsFAAOCAQEA42BdCcQLqAX+7j/dnGh6POSTdLcJHx8E
# 9LG0XIjHpf34nevqvzvoujrSWAfWd3++NpR51qaPDvBw6zZvi8+rf9AHXpHt7hfn
# 25eswIhJ0ipQKdpO3lqTPqcLVovovkqQFTNdkaYpPJLkE/HdVBzC8c76DOhHVyGE
# pMGztfsrUhsZj0mr7IGJVnquwYCHJCuVx7zJ818B3fDccdCwmkl4ZKWGpdgWpnqP
# awMCZtGMbZCrzMfxiJvF0fyvUL1G+f2QvhX71TXrUAeZnC9ETOh+G1F2PvfuXXvx
# h8kes9vFEXvpfdAAvW2fK+GFxYP84pcZ+N+Ug4MZM66kef+/Er6S9zCCBTAwggQY
# oAMCAQICEAQJGBtf1btmdVNDtW+VUAgwDQYJKoZIhvcNAQELBQAwZTELMAkGA1UE
# BhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2lj
# ZXJ0LmNvbTEkMCIGA1UEAxMbRGlnaUNlcnQgQXNzdXJlZCBJRCBSb290IENBMB4X
# DTEzMTAyMjEyMDAwMFoXDTI4MTAyMjEyMDAwMFowcjELMAkGA1UEBhMCVVMxFTAT
# BgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEx
# MC8GA1UEAxMoRGlnaUNlcnQgU0hBMiBBc3N1cmVkIElEIENvZGUgU2lnbmluZyBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAPjTsxx/DhGvZ3cH0wsx
# SRnP0PtFmbE620T1f+Wondsy13Hqdp0FLreP+pJDwKX5idQ3Gde2qvCchqXYJawO
# eSg6funRZ9PG+yknx9N7I5TkkSOWkHeC+aGEI2YSVDNQdLEoJrskacLCUvIUZ4qJ
# RdQtoaPpiCwgla4cSocI3wz14k1gGL6qxLKucDFmM3E+rHCiq85/6XzLkqHlOzEc
# z+ryCuRXu0q16XTmK/5sy350OTYNkO/ktU6kqepqCquE86xnTrXE94zRICUj6whk
# PlKWwfIPEvTFjg/BougsUfdzvL2FsWKDc0GCB+Q4i2pzINAPZHM8np+mM6n9Gd8l
# k9ECAwEAAaOCAc0wggHJMBIGA1UdEwEB/wQIMAYBAf8CAQAwDgYDVR0PAQH/BAQD
# AgGGMBMGA1UdJQQMMAoGCCsGAQUFBwMDMHkGCCsGAQUFBwEBBG0wazAkBggrBgEF
# BQcwAYYYaHR0cDovL29jc3AuZGlnaWNlcnQuY29tMEMGCCsGAQUFBzAChjdodHRw
# Oi8vY2FjZXJ0cy5kaWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURSb290Q0Eu
# Y3J0MIGBBgNVHR8EejB4MDqgOKA2hjRodHRwOi8vY3JsNC5kaWdpY2VydC5jb20v
# RGlnaUNlcnRBc3N1cmVkSURSb290Q0EuY3JsMDqgOKA2hjRodHRwOi8vY3JsMy5k
# aWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURSb290Q0EuY3JsME8GA1UdIARI
# MEYwOAYKYIZIAYb9bAACBDAqMCgGCCsGAQUFBwIBFhxodHRwczovL3d3dy5kaWdp
# Y2VydC5jb20vQ1BTMAoGCGCGSAGG/WwDMB0GA1UdDgQWBBRaxLl7KgqjpepxA8Bg
# +S32ZXUOWDAfBgNVHSMEGDAWgBRF66Kv9JLLgjEtUYunpyGd823IDzANBgkqhkiG
# 9w0BAQsFAAOCAQEAPuwNWiSz8yLRFcgsfCUpdqgdXRwtOhrE7zBh134LYP3DPQ/E
# r4v97yrfIFU3sOH20ZJ1D1G0bqWOWuJeJIFOEKTuP3GOYw4TS63XX0R58zYUBor3
# nEZOXP+QsRsHDpEV+7qvtVHCjSSuJMbHJyqhKSgaOnEoAjwukaPAJRHinBRHoXpo
# aK+bp1wgXNlxsQyPu6j4xRJon89Ay0BEpRPw5mQMJQhCMrI2iiQC/i9yfhzXSUWW
# 6Fkd6fp0ZGuy62ZD2rOwjNXpDd32ASDOmTFjPQgaGLOBm0/GkxAG/AeB+ova+YJJ
# 92JuoVP6EpQYhS6SkepobEQysmah5xikmmRR7zCCBmowggVSoAMCAQICEAMBmgI6
# /1ixa9bV6uYX8GYwDQYJKoZIhvcNAQEFBQAwYjELMAkGA1UEBhMCVVMxFTATBgNV
# BAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8G
# A1UEAxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xMB4XDTE0MTAyMjAwMDAwMFoX
# DTI0MTAyMjAwMDAwMFowRzELMAkGA1UEBhMCVVMxETAPBgNVBAoTCERpZ2lDZXJ0
# MSUwIwYDVQQDExxEaWdpQ2VydCBUaW1lc3RhbXAgUmVzcG9uZGVyMIIBIjANBgkq
# hkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAo2Rd/Hyz4II14OD2xirmSXU7zG7gU6mf
# H2RZ5nxrf2uMnVX4kuOe1VpjWwJJUNmDzm9m7t3LhelfpfnUh3SIRDsZyeX1kZ/G
# FDmsJOqoSyyRicxeKPRktlC39RKzc5YKZ6O+YZ+u8/0SeHUOplsU/UUjjoZEVX0Y
# hgWMVYd5SEb3yg6Np95OX+Koti1ZAmGIYXIYaLm4fO7m5zQvMXeBMB+7NgGN7yfj
# 95rwTDFkjePr+hmHqH7P7IwMNlt6wXq4eMfJBi5GEMiN6ARg27xzdPpO2P6qQPGy
# znBGg+naQKFZOtkVCVeZVjCT88lhzNAIzGvsYkKRrALA76TwiRGPdwIDAQABo4ID
# NTCCAzEwDgYDVR0PAQH/BAQDAgeAMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAww
# CgYIKwYBBQUHAwgwggG/BgNVHSAEggG2MIIBsjCCAaEGCWCGSAGG/WwHATCCAZIw
# KAYIKwYBBQUHAgEWHGh0dHBzOi8vd3d3LmRpZ2ljZXJ0LmNvbS9DUFMwggFkBggr
# BgEFBQcCAjCCAVYeggFSAEEAbgB5ACAAdQBzAGUAIABvAGYAIAB0AGgAaQBzACAA
# QwBlAHIAdABpAGYAaQBjAGEAdABlACAAYwBvAG4AcwB0AGkAdAB1AHQAZQBzACAA
# YQBjAGMAZQBwAHQAYQBuAGMAZQAgAG8AZgAgAHQAaABlACAARABpAGcAaQBDAGUA
# cgB0ACAAQwBQAC8AQwBQAFMAIABhAG4AZAAgAHQAaABlACAAUgBlAGwAeQBpAG4A
# ZwAgAFAAYQByAHQAeQAgAEEAZwByAGUAZQBtAGUAbgB0ACAAdwBoAGkAYwBoACAA
# bABpAG0AaQB0ACAAbABpAGEAYgBpAGwAaQB0AHkAIABhAG4AZAAgAGEAcgBlACAA
# aQBuAGMAbwByAHAAbwByAGEAdABlAGQAIABoAGUAcgBlAGkAbgAgAGIAeQAgAHIA
# ZQBmAGUAcgBlAG4AYwBlAC4wCwYJYIZIAYb9bAMVMB8GA1UdIwQYMBaAFBUAEisT
# mLKZB+0e36K+Vw0rZwLNMB0GA1UdDgQWBBRhWk0ktkkynUoqeRqDS/QeicHKfTB9
# BgNVHR8EdjB0MDigNqA0hjJodHRwOi8vY3JsMy5kaWdpY2VydC5jb20vRGlnaUNl
# cnRBc3N1cmVkSURDQS0xLmNybDA4oDagNIYyaHR0cDovL2NybDQuZGlnaWNlcnQu
# Y29tL0RpZ2lDZXJ0QXNzdXJlZElEQ0EtMS5jcmwwdwYIKwYBBQUHAQEEazBpMCQG
# CCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wQQYIKwYBBQUHMAKG
# NWh0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRENB
# LTEuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCdJX4bM02yJoFcm4bOIyAPgIfliP//
# sdRqLDHtOhcZcRfNqRu8WhY5AJ3jbITkWkD73gYBjDf6m7GdJH7+IKRXrVu3mrBg
# JuppVyFdNC8fcbCDlBkFazWQEKB7l8f2P+fiEUGmvWLZ8Cc9OB0obzpSCfDscGLT
# Ykuw4HOmksDTjjHYL+NtFxMG7uQDthSr849Dp3GdId0UyhVdkkHa+Q+B0Zl0DSbE
# Dn8btfWg8cZ3BigV6diT5VUW8LsKqxzbXEgnZsijiwoc5ZXarsQuWaBh3drzbaJh
# 6YoLbewSGL33VVRAA5Ira8JRwgpIr7DUbuD0FAo6G+OPPcqvao173NhEMIIGzTCC
# BbWgAwIBAgIQBv35A5YDreoACus/J7u6GzANBgkqhkiG9w0BAQUFADBlMQswCQYD
# VQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGln
# aWNlcnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBBc3N1cmVkIElEIFJvb3QgQ0Ew
# HhcNMDYxMTEwMDAwMDAwWhcNMjExMTEwMDAwMDAwWjBiMQswCQYDVQQGEwJVUzEV
# MBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29t
# MSEwHwYDVQQDExhEaWdpQ2VydCBBc3N1cmVkIElEIENBLTEwggEiMA0GCSqGSIb3
# DQEBAQUAA4IBDwAwggEKAoIBAQDogi2Z+crCQpWlgHNAcNKeVlRcqcTSQQaPyTP8
# TUWRXIGf7Syc+BZZ3561JBXCmLm0d0ncicQK2q/LXmvtrbBxMevPOkAMRk2T7It6
# NggDqww0/hhJgv7HxzFIgHweog+SDlDJxofrNj/YMMP/pvf7os1vcyP+rFYFkPAy
# IRaJxnCI+QWXfaPHQ90C6Ds97bFBo+0/vtuVSMTuHrPyvAwrmdDGXRJCgeGDboJz
# PyZLFJCuWWYKxI2+0s4Grq2Eb0iEm09AufFM8q+Y+/bOQF1c9qjxL6/siSLyaxhl
# scFzrdfx2M8eCnRcQrhofrfVdwonVnwPYqQ/MhRglf0HBKIJAgMBAAGjggN6MIID
# djAOBgNVHQ8BAf8EBAMCAYYwOwYDVR0lBDQwMgYIKwYBBQUHAwEGCCsGAQUFBwMC
# BggrBgEFBQcDAwYIKwYBBQUHAwQGCCsGAQUFBwMIMIIB0gYDVR0gBIIByTCCAcUw
# ggG0BgpghkgBhv1sAAEEMIIBpDA6BggrBgEFBQcCARYuaHR0cDovL3d3dy5kaWdp
# Y2VydC5jb20vc3NsLWNwcy1yZXBvc2l0b3J5Lmh0bTCCAWQGCCsGAQUFBwICMIIB
# Vh6CAVIAQQBuAHkAIAB1AHMAZQAgAG8AZgAgAHQAaABpAHMAIABDAGUAcgB0AGkA
# ZgBpAGMAYQB0AGUAIABjAG8AbgBzAHQAaQB0AHUAdABlAHMAIABhAGMAYwBlAHAA
# dABhAG4AYwBlACAAbwBmACAAdABoAGUAIABEAGkAZwBpAEMAZQByAHQAIABDAFAA
# LwBDAFAAUwAgAGEAbgBkACAAdABoAGUAIABSAGUAbAB5AGkAbgBnACAAUABhAHIA
# dAB5ACAAQQBnAHIAZQBlAG0AZQBuAHQAIAB3AGgAaQBjAGgAIABsAGkAbQBpAHQA
# IABsAGkAYQBiAGkAbABpAHQAeQAgAGEAbgBkACAAYQByAGUAIABpAG4AYwBvAHIA
# cABvAHIAYQB0AGUAZAAgAGgAZQByAGUAaQBuACAAYgB5ACAAcgBlAGYAZQByAGUA
# bgBjAGUALjALBglghkgBhv1sAxUwEgYDVR0TAQH/BAgwBgEB/wIBADB5BggrBgEF
# BQcBAQRtMGswJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0LmNvbTBD
# BggrBgEFBQcwAoY3aHR0cDovL2NhY2VydHMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0
# QXNzdXJlZElEUm9vdENBLmNydDCBgQYDVR0fBHoweDA6oDigNoY0aHR0cDovL2Ny
# bDMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9vdENBLmNybDA6oDig
# NoY0aHR0cDovL2NybDQuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9v
# dENBLmNybDAdBgNVHQ4EFgQUFQASKxOYspkH7R7for5XDStnAs0wHwYDVR0jBBgw
# FoAUReuir/SSy4IxLVGLp6chnfNtyA8wDQYJKoZIhvcNAQEFBQADggEBAEZQPsm3
# KCSnOB22WymvUs9S6TFHq1Zce9UNC0Gz7+x1H3Q48rJcYaKclcNQ5IK5I9G6OoZy
# rTh4rHVdFxc0ckeFlFbR67s2hHfMJKXzBBlVqefj56tizfuLLZDCwNK1lL1eT7EF
# 0g49GqkUW6aGMWKoqDPkmzmnxPXOHXh2lCVz5Cqrz5x2S+1fwksW5EtwTACJHvzF
# ebxMElf+X+EevAJdqP77BzhPDcZdkbkPZ0XN1oPt55INjbFpjE/7WeAjD9KqrgB8
# 7pxCDs+R1ye3Fu4Pw718CqDuLAhVhSK46xgaTfwqIa1JMYNHlXdx3LEbS0scEJx3
# FMGdTy9alQgpECYxggQ7MIIENwIBATCBhjByMQswCQYDVQQGEwJVUzEVMBMGA1UE
# ChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYD
# VQQDEyhEaWdpQ2VydCBTSEEyIEFzc3VyZWQgSUQgQ29kZSBTaWduaW5nIENBAhAB
# bOl77YhPJjl7V/MavkwsMAkGBSsOAwIaBQCgeDAYBgorBgEEAYI3AgEMMQowCKAC
# gAChAoAAMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQBgjcCAQsx
# DjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBTnPxDkyuTH9gTBDVChMvaE
# f/rzVzANBgkqhkiG9w0BAQEFAASCAQBy/C6LceEg5hEeGIgjpEJEwxMiZIl8bnyg
# x9DED8TIlyHcbR/Nv2lMsc/Yjose5LacrGBnZeBNCqKqp7L1v4lyWZksQbFErZUq
# jW4saThMOYkkoslMnDjI16/aCGzw0u4870t1MM1ajPlx3r/pUiKH75gA1F0DM0xY
# HQHRyq/i6pTjOnCjxdlRgqAaoLRjHDRdEEX8/oPQGLQnqQUxOG2hC12AcFJyJgV2
# pJhUDvYGvRE4XJO1GAMTFmXv65uIMKFIdjxTw8Nu1OYnIqUhtOyILgp+TG4voTTf
# K5OiG7Za0VTWq/Rv82VryK7BcCcwamrYpXVznS+YEVGUyNBWv52coYICDzCCAgsG
# CSqGSIb3DQEJBjGCAfwwggH4AgEBMHYwYjELMAkGA1UEBhMCVVMxFTATBgNVBAoT
# DERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8GA1UE
# AxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xAhADAZoCOv9YsWvW1ermF/BmMAkG
# BSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJ
# BTEPFw0xODAyMjQwNjAwMjBaMCMGCSqGSIb3DQEJBDEWBBRIlhvuZY+OEEfx9g8A
# Q9X4HLL1vjANBgkqhkiG9w0BAQEFAASCAQCIUIWIxIBEDKOZSH1scOWHJxvxqs1t
# nzoqp8Qfp77GXWxcgcIiyzK4LAMU/O2UEbUNOS9wRaBjujd463IFWVoSTSVvhuRf
# H2Iok5kOZHzvmBS8U9T/UR4ckIkNcpvIBThGuyNid/tW2/k2EkO4S2CV9Z6E3vB4
# g+AN3Ggw/KZGAgZvQ8BQqVADu6JrrLOqI997NGlkbCV95MX9aZhpQtyNeWbH5oU3
# P1WfI4cuQ3CyAvyuNS4CSX/Cqv+tenOpgJdPML+BemdH+K9qtAiVHOuZhJT2vXMq
# k36ghTtGKJOHTEricOW/2i7vwntSY4HHDpAKzH7xfms2SumfJ8u7wzO8
# SIG # End signature block
